/* This file is part Tempura: Interval Temporal Logic interpreter.
 *  
 * Copyright (C) 1987-2021, 2023 Roger Hale, Ben Moszkowski, Antonio Cau 
 *
 * Tempura is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 *
 * Tempura is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with Tempura.  If not, see <http://www.gnu.org/licenses/>.
 *
 * Changes made by  AC: Antonio Cau
 *                 BCM: Ben Moszkowski
 *
 * $Id: top-lev.c 285 2017-10-08 08:25:03Z cau $
 * 									 
 *  History:								 
 * 									 
 *   The first Tempura interpreter was programmed in Prolog by Ben	 
 *   Moszkowski, and was operational around December 2, 1983.  Subsequently
 *   he rewrote the interpreter in Lisp (mid Mar, 1984), and in late 1984  
 *   modified the program to handle a two-level memory and multi-pass	   
 *   scanning. The C-Tempura interpreter was written in early 1985 by	   
 *   Roger Hale at Cambridge University.				  
 * 									  
 * Some known bugs to fix:						  
 *   "struct" is rather flaky, but believed to be OK if used as		  
 *   "stable(struct(L))".
 */


#include "system.h"



/*************************************\
* 				      *
* top-level read-eval-print routines  *
* 				      *
\*************************************/


/* C top level */
int main (int argc, char **argv)
{
    tempura_init ();
    printf("Tempura: Interval Temporal Logic interpreter\n\n");
    fflush(stdout);
    printf("Copyright (C) 1987-2023  Roger Hale, Ben Moszkowski, Antonio Cau\n\n");
    fflush(stdout);
    printf("Tempura is free software: you can redistribute it and/or modify\n");
    fflush(stdout);
    printf("it under the terms of the GNU General Public License as published by\n");
    fflush(stdout);
    printf("the Free Software Foundation, either version 3 of the License, or\n");
    fflush(stdout);
    printf("(at your option) any later version.\n\n");
    fflush(stdout);
    printf("Tempura is distributed in the hope that it will be useful,\n");
    fflush(stdout);
    printf("but WITHOUT ANY WARRANTY; without even the implied warranty of\n");
    fflush(stdout);
    printf("MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n");
    fflush(stdout);
    printf("GNU General Public License for more details.\n\n");
    fflush(stdout);
    printf("You should have received a copy of the GNU General Public License\n");
    fflush(stdout);
    printf("along with Tempura.  If not, see <http://www.gnu.org/licenses/>.\n\n");
    fflush(stdout);
    printf("Current maintainers:\n");
    fflush(stdout);
    printf("  Antonio Cau and Ben Moszkowski, (cau.researcher@gmail.com)\n\n");
    fflush(stdout);
    printf ("C-Tempura (Version ");
    fflush(stdout);
    print_version_number();
    printf(")\n\n");
    fflush(stdout);

    tempura (TempuraLevel = 0);
    if (Fpcur != NULL)
        fclose(Fpcur);
    exit(0);
}

void prompt (int level)
{
    int     i;
    printf ("Tempura %d", Counter++);
    for (i = 0; i <= level; i++)
    putchar ('%');
    fputs (" ", stdout);
    newline;
    fflush(stdout);
}

/* main read_run loop */
void tempura (int level)
{   
    int val;
    jmp_buf *OldBreakPoint= BreakPoint;
    save_state (BreakPoint);
    while (TempuraLevel >= level) {
        prompt (level);
        INTERACTIVE = 1;yylineno=1;
        Linetext[0]='\0';plinetext=Linetext;yycolnum = 1;
        exec_command (LastCommand = parse (stdin));
    }
    BreakPoint = OldBreakPoint;
    if (Break)
        restore_state (BreakPoint,val);
}


/***********************\
*                      *
*  command interpreter *
*                      *
\***********************/

void exec_command (NODEPTR cmd)
{
    BreakDepth = -1;
    cmd = reduce_stmt (cmd);
    assert0 (cmd != False, "the command is false!");
    if (!Ready) {
        printf("Unable to evaluate command:\n");
        output_expn(cmd,-1);
        newline;
    }
    assert0 (Ready, "unable to evaluate top-level command in one pass");
    
    
}

/*********************\
*                    *
* reduction routines *
* 	             *
\*********************/






#define applystmt(R) ( (void (*)(NODEPTR)) (R[token_type(node_type(ReduceStackTop -> n))][0]))(ReduceStackTop -> n)
#define applyval(R) ( (void (*)(NODEPTR)) (R[token_type(node_type(ReduceStackTop -> n))][1]))(ReduceStackTop -> n)
#define applyref(R) ( (void (*)(NODEPTR)) (R[token_type(node_type(ReduceStackTop -> n))][2]))(ReduceStackTop -> n)

void out_reduction(char *s)
{
    int n=22-strlen(s);

    printf("----");
    while (n-- > 0)
        printf("-");
    printf("[%s]---->\n",s);
}

    

/* reduce a formula */
NODEPTR reduce_stmt (NODEPTR e)
{
    push_expn(e);
    TempNodeList = NULL;   /* This was moved from within loop. BCM 25Jul97 */
    TempNodes = 0;
    do {
        if (InterruptSet)
            user_break ("interrupt");
        Retry = F;
        if (ReduceDepth <= BreakDepth) {
            show_form ();
            tempura (++TempuraLevel);
        }
            /* TempNodeList = NULL;*/
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
            printf("(%4d;%4d) R_stmt  pre: ",StateNumber,ReduceDepth);
            output_expn(ReduceStackTop -> n,PrintLevel);
            printf("\n");
            out_reduction(Reduc[token_type(node_type(ReduceStackTop -> n))][0]);
            fflush(stdout);
	};
                
        applystmt (Reduces);
        
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
            printf("(%4d;%4d) E=",StateNumber,ReduceDepth);
            output_expn(node_next(EmptyLoc),-1);
            printf(" FO=%1d, SPF=%1d, PPF=%1d, Rdy=%1d, Rtry=%1d, R_stmt post: ",ForceOutput,StateProgressFlag,PassProgressFlag,Ready,Retry);
            output_expn(ReduceStackTop -> n,PrintLevel);           
            printf("\n");
            fflush(stdout);
	};
        
    } while (Retry);
    if (Ready) {
        PassSuccessfulReduceCount += 1;
        PassProgressFlag = T;
    }
    else if (PassLimitCounter == PassLimit) {
        PassLimitCounter = 0;
        if (DebugLevel>=5) 
            printf("Reached PassLimit\n");
	state_not_defined ();
    }
    else{
        
        if (!StateProgressFlag){
            if (DebugLevel>=5)
                printf("No state progress\n");
            state_not_defined ();
        }
        
        
    }
    

    return (pop_expn);
}

/* reduce a value */
NODEPTR reduce_val (NODEPTR e)
{
    push_expn (e);
    do {
        Retry = F;
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
            printf("(%4d,%4d) R_val   pre: ",StateNumber,ReduceDepth);
            output_expn(ReduceStackTop -> n,PrintLevel);
            printf("\n");
            out_reduction(Reduc[token_type(node_type(ReduceStackTop -> n))][1]);
            fflush(stdout);
	};
        applyval (Reduces);
        
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
            printf("(%4d;%4d) E=",StateNumber,ReduceDepth);
            output_expn(node_next(EmptyLoc),-1);
            printf(" FO=%1d, SPF=%1d, PPF=%1d, Rdy=%1d, Rtry=%1d, R_val  post: ",ForceOutput,StateProgressFlag,PassProgressFlag,Ready,Retry);
          output_expn(ReduceStackTop -> n,PrintLevel);
          printf("\n");
          fflush(stdout);
	};
    } while (Retry);

    if (Ready)
        PassSuccessfulReduceCount += 1;
    return (pop_expn);
}

/* reduce a reference */
NODEPTR reduce_ref (NODEPTR e)
{
    push_expn (e);
    do {
        Retry = F;
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
            printf("(%4d,%4d) R_ref   pre: ",StateNumber,ReduceDepth);
            output_expn(ReduceStackTop -> n,PrintLevel);
	    printf("\n");       
            out_reduction(Reduc[token_type(node_type(ReduceStackTop -> n))][2]);
            fflush(stdout);
	};
        
        applyref (Reduces);
        if (DebugLevel>=5) {  /* DEBUG  BCM 29Jul97 */
            printf("(%4d;%4d) E=",StateNumber,ReduceDepth);
            output_expn(node_next(EmptyLoc),-1);
            printf(" FO=%1d, SPF=%1d, PPF=%1d, Rdy=%1d, Rtry=%1d, R_ref  post: ",ForceOutput,StateProgressFlag,PassProgressFlag,Ready,Retry);
	  output_expn(ReduceStackTop -> n,PrintLevel);
          printf("\n");
          fflush(stdout);
	};
        
    } while (Retry);

    if (Ready)
        PassSuccessfulReduceCount += 1;
    return (pop_ref);
}



/***********************************************\
* 	                                       *
*  reduce a subprogram with its own empty-flag *
* 		                               *
\***********************************************/


NODEPTR reduce_pgm (NODEPTR e)
{
    e = reduce_stmt (e);
    if (Ready) {
        assert0 (e != False, "the program is false!");
        assert0 (empty_defined (), "the interval length is undefined");
    }
    return (e);
}

/*******************************************\
*                                           *
*  Reduce a reference and fetch its value.  *
* 		                            *
\*******************************************/
NODEPTR reduce_ref_and_fetch1(NODEPTR e)
{
    NODEPTR loc = reduce_ref (e);
    NODEPTR ploc;
    NODEPTR val;
    unsigned int k;
    
    /*printf("loc=%p md=%d is_cur=%d\n",loc, mem_defined1(loc),is_curr(loc));*/
    
    if (Ready) {
            if ((Ready = mem_defined1 (loc))){
                if (is_curr(loc)) {
                    val = get_mem(loc);
                    /*output_expn(val, -1);newline;
                    show_mem_all1(loc);
                    printf("curr_val_defined=%d\n",curr_val_defined(val));*/
                    if (curr_val_defined(val)) {
                        e = copy_curr_val(val);
                    }
                    else {
                        e = val;
                    }
                }

                if (is_next(loc)) {
                    val = get_mem(loc);
                    if (next_val_defined(val)) {
                        e = copy_next_val(val);
                    }
                    else {
                        e = val;
                    }
                }
                for(k=2;k<=1+node_mem_history(loc);k = k+1) {
                    if (is_prev1(loc,k)) {
                        val= get_mem(loc);
                        if (prev_val_defined(val,k)) {
                            e = copy_prev_val(val,k);
                        }
                        else {
                            e = val;
                        }
                    }
                }
            } 
        }
    return (e);
}



NODEPTR reduce_ref_and_fetch(NODEPTR e)
{
    NODEPTR loc = reduce_ref (e);
    NODEPTR ploc;
    NODEPTR val;
        
    
    if (Ready) {
        if ((Ready = mem_defined1 (loc))){
            e = get_mem(loc);
        }
    }
    return (e);
}

unsigned int reduce_ref_and_fetch_history(NODEPTR e)
{
    NODEPTR loc = reduce_ref (e);
    unsigned int history = 0;
    
    if (Ready) {
        if ((Ready = mem_defined1 (loc)))
            history = node_mem_history (loc);
    }
    return (history);
}

unsigned reduce_ref_and_fetch_is_undefined(NODEPTR e)
{
    NODEPTR loc = reduce_ref (e);
    unsigned und =0;
    
    if (Ready) {
            if ((Ready = mem_defined1 (loc)))
                und = node_mem_isundefined (loc);
        }
    return (und);
}

unsigned reduce_ref_and_fetch_has_default(NODEPTR e)
{
    NODEPTR loc = reduce_ref (e);
    unsigned hasd =0;
    
    if (Ready) {
        if ((Ready = mem_defined1 (loc)))
            hasd = node_mem_hasdefault (loc);
    }
    return (hasd);
}

/* Convert a value to a reference. */
NODEPTR make_ref (NODEPTR e)
{
    TempLoc = adj_loc1 (TempLoc);
    put_mem1 (TempLoc, e);
    return (TempLoc);
}

void get_val(NODEPTR e)
{
    NODEPTR val = reduce_ref_and_fetch (e);
    if (Ready) {
        if ((Ready = val_defined (val))) 
            result (val);
    }
}

void get_ref(NODEPTR e)
{
    NODEPTR val = reduce_val (e);
    if (Ready)
        ref_result (make_ref (val));
}



/* end of top-level.c */

